@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %% 
-%% Copyright Ericsson AB 2000-2013. All Rights Reserved.
+%% Copyright Ericsson AB 2000-2019. All Rights Reserved.
 %% 
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -29,9 +29,9 @@
 -export([open/3, open/4, fdopen/4, fdopen/5, close/1]).
 -export([bind/3, listen/1, listen/2, peeloff/2]).
 -export([connect/3, connect/4, async_connect/4]).
--export([accept/1, accept/2, async_accept/2]).
+-export([accept/1, accept/2, accept/3, async_accept/2]).
 -export([shutdown/2]).
--export([send/2, send/3, sendto/4, sendmsg/3]).
+-export([send/2, send/3, sendto/4, sendmsg/3, sendfile/4]).
 -export([recv/2, recv/3, async_recv/3]).
 -export([unrecv/2]).
 -export([recvfrom/2, recvfrom/3]).
@@ -49,9 +49,15 @@
 -include("inet_sctp.hrl").
 -include("inet_int.hrl").
 
-%-define(DEBUG, 1).
+%%%-define(DEBUG, 1).
 -ifdef(DEBUG).
--define(DBG_FORMAT(Format, Args), (io:format((Format), (Args)))).
+-define(
+   DBG_FORMAT(Format, Args),
+   begin
+       %% io:format((Format), (Args)),
+       erlang:display(lists:flatten(io_lib:format((Format), (Args)))),
+       ok
+   end).
 -else.
 -define(DBG_FORMAT(Format, Args), ok).
 -endif.
@@ -70,11 +76,13 @@ open(Protocol, Family, Type) ->
 open(Protocol, Family, Type, Opts) ->
     open(Protocol, Family, Type, Opts, ?INET_REQ_OPEN, []).
 
+%% FDOPEN(tcp|udp|sctp, inet|inet6|local, stream|dgram|seqpacket, integer())
+
 fdopen(Protocol, Family, Type, Fd) when is_integer(Fd) ->
     fdopen(Protocol, Family, Type, Fd, true).
 
 fdopen(Protocol, Family, Type, Fd, Bound)
-  when is_integer(Fd), Bound == true orelse Bound == false ->
+  when is_integer(Fd), is_boolean(Bound) ->
     open(Protocol, Family, Type, [], ?INET_REQ_FDOPEN,
          [?int32(Fd), enc_value_2(bool, Bound)]).
 
@@ -104,8 +112,9 @@ open(Protocol, Family, Type, Opts, Req, Data) ->
 	error:system_limit -> {error, system_limit}
     end.
 
-enc_family(inet) -> ?INET_AF_INET;
-enc_family(inet6) -> ?INET_AF_INET6.
+enc_family(inet)  -> ?INET_AF_INET;
+enc_family(inet6) -> ?INET_AF_INET6;
+enc_family(local) -> ?INET_AF_LOCAL.
 
 enc_type(stream) -> ?INET_TYPE_STREAM;
 enc_type(dgram) -> ?INET_TYPE_DGRAM;
@@ -147,39 +156,106 @@ shutdown_1(S, How) ->
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 close(S) when is_port(S) ->
+    ?DBG_FORMAT("prim_inet:close(~p)~n", [S]),
     case getopt(S, linger) of
     	{ok,{true,0}} -> 
 	    close_port(S);
-	_ ->
-	    case subscribe(S, [subs_empty_out_q]) of
-		{ok, [{subs_empty_out_q,N}]} when N > 0 ->
-		    close_pend_loop(S, N);   %% wait for pending output to be sent
-		_ ->
-		    close_port(S)
-	    end
+        {ok,{true,T}} ->
+            %% Wait for T seconds for pending output to be sent
+            %%
+            %% Note that this handling of Linger may look ok,
+            %% but sweeps some problems under the rug since
+            %% there are OS buffers that may have remaining data
+            %% after the inet driver has emptied its buffers.
+            %% But Linger for nonblocking sockets is broken
+            %% anyway on all OS:es, according to hearsay,
+            %% and is a contradiction in itself.
+            %% We have hereby done our best...
+            %%
+            case subscribe(S, [subs_empty_out_q]) of
+                {ok, [{subs_empty_out_q,0}]} ->
+                    close_port(S);
+                {ok, [{subs_empty_out_q,N}]} when N > 0 ->
+                    %% Wait for pending output to be sent
+                    Tref = erlang:start_timer(T * 1000, self(), close_port),
+                    close_pend_loop(S, Tref, N);
+                _ ->
+                    %% Subscribe failed - wait full time
+                    Tref = erlang:start_timer(T * 1000, self(), close_port),
+                    close_pend_loop(S, Tref, undefined)
+            end;
+	_ -> % Regard this as {ok,{false,_}}
+            case subscribe(S, [subs_empty_out_q]) of
+                {ok, [{subs_empty_out_q,N}]} when N > 0 ->
+                    %% Wait for pending output to be sent
+                    DefaultT = 180000, % Arbitrary system timeout 3 min
+                    Tref = erlang:start_timer(DefaultT, self(), close_port),
+                    close_pend_loop(S, Tref, N);
+                _ ->
+                    %% Subscribe failed or empty out q - give up or done
+                    close_port(S)
+            end
     end.
 
-close_pend_loop(S, N) ->
+close_pend_loop(S, Tref, N) ->
+    ?DBG_FORMAT("prim_inet:close_pend_loop(~p, _, ~p)~n", [S,N]),
     receive
-	{empty_out_q,S} ->
-	    close_port(S)
+        {timeout,Tref,_} -> % Linger timeout
+            ?DBG_FORMAT("prim_inet:close_pend_loop(~p, _, _) timeout~n", [S]),
+	    close_port(S);
+	{empty_out_q,S} when N =/= undefined ->
+            ?DBG_FORMAT(
+               "prim_inet:close_pend_loop(~p, _, _) empty_out_q~n", [S]),
+	    close_port(S, Tref)
     after ?INET_CLOSE_TIMEOUT ->
 	    case getstat(S, [send_pend]) of
                 {ok, [{send_pend,N1}]} ->
+                    ?DBG_FORMAT(
+                       "prim_inet:close_pend_loop(~p, _, _) send_pend ~p~n",
+                       [S,N1]),
                     if
-			N1 =:= N ->
-			    close_port(S);
-                       true ->
-			    close_pend_loop(S, N1)
+                        N1 =:= 0 ->
+                            %% Empty outq - done
+                            close_port(S, Tref);
+                        N =:= undefined ->
+                            %% Within linger time - wait some more
+                            close_pend_loop(S, Tref, N);
+                        N1 =:= N ->
+                            %% Inactivity - give up
+                            close_port(S, Tref);
+                        true ->
+                            %% Still moving - wait some more
+                            close_pend_loop(S, Tref, N)
                     end;
-		_ ->
-		    close_port(S)
-	    end
+                _Stat ->
+                    %% Failed getstat - give up
+                    ?DBG_FORMAT(
+                       "prim_inet:close_pend_loop(~p, _, _) getstat ~p~n",
+                       [S,_Stat]),
+		    close_port(S, Tref)
+            end
     end.
 
+
+close_port(S, Tref) ->
+    ?DBG_FORMAT("prim_inet:close_port(~p, _)~n", [S]),
+    case erlang:cancel_timer(Tref) of
+        false ->
+            receive
+                {timeout,Tref,_} ->
+                    ok
+            end;
+        _N ->
+            ok
+    end,
+    close_port(S).
+%%
 close_port(S) ->
-    catch erlang:port_close(S),
-    receive {'EXIT',S,_} -> ok after 0 -> ok end.
+    ?DBG_FORMAT("prim_inet:close_port(~p)~n", [S]),
+    _Closed = (catch erlang:port_close(S)),
+    receive {'EXIT',S,_} -> ok after 0 -> ok end,
+    ?DBG_FORMAT("prim_inet:close_port(~p) ~p~n", [S,_Closed]),
+    ok.
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%
@@ -189,41 +265,52 @@ close_port(S) ->
 %%
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
-bind(S,IP,Port) when is_port(S), is_integer(Port), Port >= 0, Port =< 65535 ->
-    case ctl_cmd(S,?INET_REQ_BIND,enc_value(set, addr, {IP,Port})) of
-	{ok, [P1,P0]} -> {ok, ?u16(P1, P0)};
-	{error,_}=Error -> Error
-    end;
-
 %% Multi-homed "bind": sctp_bindx(). The Op is 'add' or 'remove'.
 %% If no addrs are specified, it just does nothing.
 %% Function returns {ok, S} on success, unlike TCP/UDP "bind":
-bind(S, Op, Addrs) when is_port(S), is_list(Addrs) ->
-    case Op of
-	add ->
-	    bindx(S, 1, Addrs);
-	remove ->
-	    bindx(S, 0, Addrs);
-	_ -> {error, einval}
+bind(S, add, Addrs) when is_port(S), is_list(Addrs) ->
+    bindx(S, 1, Addrs);
+bind(S, remove, Addrs) when is_port(S), is_list(Addrs) ->
+    bindx(S, 0, Addrs);
+bind(S, Addr, _) when is_port(S), tuple_size(Addr) =:= 2 ->
+    case type_value(set, addr, Addr) of
+	true ->
+	    case ctl_cmd(S,?INET_REQ_BIND,enc_value(set, addr, Addr)) of
+		{ok, [P1,P0]} -> {ok, ?u16(P1, P0)};
+		{error, _} = Error -> Error
+	    end;
+	false ->
+	    {error, einval}
     end;
-bind(_, _, _) -> {error, einval}.
+bind(S, IP, Port) ->
+    bind(S, {IP, Port}, 0).
 
 bindx(S, AddFlag, Addrs) ->
     case getprotocol(S) of
 	sctp ->
-	    %% Really multi-homed "bindx". Stringified args:
-	    %% [AddFlag, (AddrBytes see enc_value_2(addr,X))+]:
-	    Args =
-		[?int8(AddFlag)|
-		 [enc_value(set, addr, {IP,Port}) ||
-		     {IP, Port} <- Addrs]],
-	    case ctl_cmd(S, ?SCTP_REQ_BINDX, Args) of
-		{ok,_} -> {ok, S};
-		{error,_}=Error  -> Error
+	    case bindx_check_addrs(Addrs) of
+		true ->
+		    %% Really multi-homed "bindx". Stringified args:
+		    %% [AddFlag, (AddrBytes see enc_value_2(addr,X))+]:
+		    Args =
+			[?int8(AddFlag)|
+			 [enc_value(set, addr, Addr) || Addr <- Addrs]],
+		    case ctl_cmd(S, ?SCTP_REQ_BINDX, Args) of
+			{ok, _} -> {ok, S};
+			{error, _}=Error  -> Error
+		    end;
+		false ->
+		    {error, einval}
 	    end;
-	_ -> {error, einval}
+	_ ->
+	    {error, einval}
     end.
 
+bindx_check_addrs([Addr|Addrs]) ->
+    type_value(set, addr, Addr) andalso bindx_check_addrs(Addrs);
+bindx_check_addrs([]) ->
+    true.
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%
 %% CONNECT(insock(), IP, Port [,Timeout]) -> ok | {error, Reason}
@@ -232,7 +319,7 @@ bindx(S, AddFlag, Addrs) ->
 %% if timeout is given:
 %%       timeout < 0  -> infinity
 %%                 0  -> immediate connect (mostly works for loopback)
-%%               > 0  -> wait for timout ms if not connected then 
+%%               > 0  -> wait for timeout ms if not connected then 
 %%                       return {error, timeout} 
 %%
 %% ASYNC_CONNECT(insock(), IP, Port, Timeout) -> {ok, S, Ref} | {error, Reason}
@@ -242,14 +329,24 @@ bindx(S, AddFlag, Addrs) ->
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% For TCP, UDP or SCTP sockets.
 %%
-connect(S, IP, Port) -> connect0(S, IP, Port, -1).
 
-connect(S, IP, Port, infinity) -> connect0(S, IP, Port, -1);
-connect(S, IP, Port, Time)     -> connect0(S, IP, Port, Time).
+connect(S, IP, Port) ->
+    connect(S, IP, Port, infinity).
+%%
+connect(S, Addr, _, Time) when is_port(S), tuple_size(Addr) =:= 2 ->
+    case type_value(set, addr, Addr) of
+	true when Time =:= infinity ->
+	    connect0(S, Addr, -1);
+	true when is_integer(Time) ->
+	    connect0(S, Addr, Time);
+	false ->
+	    {error, einval}
+    end;
+connect(S, IP, Port, Time) ->
+    connect(S, {IP, Port}, 0, Time).
 
-connect0(S, IP, Port, Time) when is_port(S), Port > 0, Port =< 65535,
-				 is_integer(Time) ->
-    case async_connect(S, IP, Port, Time) of
+connect0(S, Addr, Time) ->
+    case async_connect0(S, Addr, Time) of
 	{ok, S, Ref} ->
 	    receive
 		{inet_async, S, Ref, Status} ->
@@ -258,23 +355,41 @@ connect0(S, IP, Port, Time) when is_port(S), Port > 0, Port =< 65535,
 	Error -> Error
     end.
 
+
+async_connect(S, Addr, _, Time) when is_port(S), tuple_size(Addr) =:= 2 ->
+    case type_value(set, addr, Addr) of
+	true when Time =:= infinity ->
+	    async_connect0(S, Addr, -1);
+	true when is_integer(Time) ->
+	    async_connect0(S, Addr, Time);
+	false ->
+	    {error, einval}
+    end;
+%%
 async_connect(S, IP, Port, Time) ->
-    case ctl_cmd(S, ?INET_REQ_CONNECT,
-		 [enc_time(Time),?int16(Port),ip_to_bytes(IP)]) of
+    async_connect(S, {IP, Port}, 0, Time).
+
+async_connect0(S, Addr, Time) ->
+    case ctl_cmd(
+	   S, ?INET_REQ_CONNECT,
+	   [enc_time(Time),enc_value(set, addr, Addr)])
+    of
 	{ok, [R1,R0]} -> {ok, S, ?u16(R1,R0)};
-	{error,_}=Error -> Error
+	{error, _}=Error -> Error
     end.
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%
-%% ACCEPT(insock() [,Timeout] ) -> {ok,insock()} | {error, Reason}
+%% ACCEPT(insock() [,Timeout][,FamilyOpts] ) -> {ok,insock()} | {error, Reason}
 %%
 %% accept incoming connection on listen socket 
 %% if timeout is given:
 %%       timeout < 0  -> infinity
 %%                 0  -> immediate accept (poll)
-%%               > 0  -> wait for timout ms for accept if no accept then 
+%%               > 0  -> wait for timeout ms for accept if no accept then 
 %%                       return {error, timeout}
+%% FamilyOpts are address family specific options to copy from
+%% listen socket to accepted socket
 %%
 %% ASYNC_ACCEPT(insock(), Timeout)
 %%
@@ -285,17 +400,22 @@ async_connect(S, IP, Port, Time) ->
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% For TCP sockets only.
 %%
-accept(L)            -> accept0(L, -1).
+accept(L) -> accept0(L, -1, []).
 
-accept(L, infinity)  -> accept0(L, -1);
-accept(L, Time)      -> accept0(L, Time).
+accept(L, infinity) -> accept0(L, -1, []);
+accept(L, FamilyOpts) when is_list(FamilyOpts) -> accept0(L, -1, FamilyOpts);
+accept(L, Time) -> accept0(L, Time, []).
 
-accept0(L, Time) when is_port(L), is_integer(Time) ->
+accept(L, infinity, FamilyOpts) -> accept0(L, -1, FamilyOpts);
+accept(L, Time, FamilyOpts) -> accept0(L, Time, FamilyOpts).
+
+accept0(L, Time, FamilyOpts)
+  when is_port(L), is_integer(Time), is_list(FamilyOpts) ->
     case async_accept(L, Time) of
 	{ok, Ref} ->
 	    receive 
 		{inet_async, L, Ref, {ok,S}} ->
-		    accept_opts(L, S);
+		    accept_opts(L, S, FamilyOpts);
 		{inet_async, L, Ref, Error} ->
 		    Error
 	    end;
@@ -303,15 +423,22 @@ accept0(L, Time) when is_port(L), is_integer(Time) ->
     end.
 
 %% setup options from listen socket on the connected socket
-accept_opts(L, S) ->
-    case getopts(L, [active, nodelay, keepalive, delay_send, priority, tos]) of
+accept_opts(L, S, FamilyOpts) ->
+    case
+        getopts(
+          L,
+          [active, nodelay, keepalive, delay_send, priority]
+          ++ FamilyOpts)
+    of
 	{ok, Opts} ->
-	    case setopts(S, Opts) of
-		ok -> {ok, S};
-		Error -> close(S), Error
-	    end;
-	Error ->
-	    close(S), Error
+            case setopts(S, Opts) of
+                ok ->
+                    {ok, S};
+                Error1 ->
+                    close(S), Error1
+            end;
+	Error2 ->
+	    close(S), Error2
     end.
 
 async_accept(L, Time) ->
@@ -370,23 +497,49 @@ peeloff(S, AssocId) ->
 %% be called directly -- use "sendmsg" instead:
 %%
 send(S, Data, OptList) when is_port(S), is_list(OptList) ->
-    ?DBG_FORMAT("prim_inet:send(~p, ~p)~n", [S,Data]),
+    ?DBG_FORMAT("prim_inet:send(~p, _, ~p)~n", [S,OptList]),
     try erlang:port_command(S, Data, OptList) of
 	false -> % Port busy and nosuspend option passed
 	    ?DBG_FORMAT("prim_inet:send() -> {error,busy}~n", []),
 	    {error,busy};
 	true ->
-	    receive
-		{inet_reply,S,Status} ->
-		    ?DBG_FORMAT("prim_inet:send() -> ~p~n", [Status]),
-		    Status
-	    end
+            send_recv_reply(S, undefined)
     catch
 	error:_Error ->
 	    ?DBG_FORMAT("prim_inet:send() -> {error,einval}~n", []),
 	     {error,einval}
     end.
 
+send_recv_reply(S, Mref) ->
+    ReplyTimeout =
+        case Mref of
+            undefined ->
+                ?INET_CLOSE_TIMEOUT;
+            _ ->
+                infinity
+        end,
+    receive
+        {inet_reply,S,Status} ->
+            ?DBG_FORMAT(
+               "prim_inet:send_recv_reply(~p, _): inet_reply ~p~n",
+               [S,Status]),
+            case Mref of
+                undefined -> ok;
+                _ ->
+                    demonitor(Mref, [flush]),
+                    ok
+            end,
+            Status;
+        {'DOWN',Mref,_,_,_Reason} when Mref =/= undefined ->
+            ?DBG_FORMAT(
+               "prim_inet:send_recv_reply(~p, _) 'DOWN' ~p~n",
+               [S,_Reason]),
+            {error,closed}
+    after ReplyTimeout ->
+            send_recv_reply(S, monitor(port, S))
+    end.
+
+
 send(S, Data) ->
     send(S, Data, []).
 
@@ -400,20 +553,34 @@ send(S, Data) ->
 %% "sendto" is for UDP. IP and Port are set by the caller to 0 if the socket
 %% is known to be connected.
 
-sendto(S, IP, Port, Data) when is_port(S), Port >= 0, Port =< 65535 ->
-    ?DBG_FORMAT("prim_inet:sendto(~p, ~p, ~p, ~p)~n", [S,IP,Port,Data]),
-    try erlang:port_command(S, [?int16(Port),ip_to_bytes(IP),Data]) of
-	true -> 
-	    receive
-		{inet_reply,S,Reply} ->
-		    ?DBG_FORMAT("prim_inet:sendto() -> ~p~n", [Reply]),
-		     Reply
-	    end
-    catch
-	error:_ ->
-	    ?DBG_FORMAT("prim_inet:sendto() -> {error,einval}~n", []),
-	     {error,einval}
-    end.
+sendto(S, Addr, _, Data) when is_port(S), tuple_size(Addr) =:= 2 ->
+    case type_value(set, addr, Addr) of
+	true ->
+	    ?DBG_FORMAT("prim_inet:sendto(~p, ~p, ~p)~n", [S,Addr,Data]),
+	    try
+		erlang:port_command(S, [enc_value(set, addr, Addr),Data])
+	    of
+		true ->
+		    receive
+			{inet_reply,S,Reply} ->
+			    ?DBG_FORMAT(
+			       "prim_inet:sendto() -> ~p~n", [Reply]),
+			    Reply
+		    end
+	    catch
+		error:_ ->
+		    ?DBG_FORMAT(
+		       "prim_inet:sendto() -> {error,einval}~n", []),
+		    {error,einval}
+	    end;
+	false ->
+	    ?DBG_FORMAT(
+	       "prim_inet:sendto() -> {error,einval}~n", []),
+	    {error,einval}
+    end;
+sendto(S, IP, Port, Data) ->
+    sendto(S, {IP, Port}, 0, Data).
+
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%
@@ -434,6 +601,77 @@ sendmsg(S, #sctp_sndrcvinfo{}=SRI, Data) when is_port(S) ->
 	Reason -> {error,Reason}
     end.
 
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%%
+%% SENDFILE(outsock(), Fd, Offset, Length) -> {ok,BytesSent} | {error, Reason}
+%%
+%% send Length data bytes from a file handle, to a socket, starting at Offset
+%%
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%% "sendfile" is for TCP:
+
+sendfile(S, FileHandle, Offset, Length)
+        when not is_port(S);
+             not is_binary(FileHandle);
+             not is_integer(Offset);
+             not is_integer(Length) ->
+    {error, badarg};
+sendfile(S, FileHandle, Offset, Length) ->
+    case erlang:port_info(S, connected) of
+        {connected, Pid} when Pid =:= self() ->
+            Uncork = sendfile_maybe_cork(S),
+            Result = sendfile_1(S, FileHandle, Offset, Length),
+            sendfile_maybe_uncork(S, Uncork),
+            Result;
+        {connected, Pid} when Pid =/= self() ->
+            {error, not_owner};
+        _Other ->
+            {error, einval}
+    end.
+
+sendfile_maybe_cork(S) ->
+    case getprotocol(S) of
+        tcp ->
+            case getopts(S, [nopush]) of
+                {ok, [{nopush,false}]} ->
+                    _ = setopts(S, [{nopush,true}]),
+                    true;
+                _ ->
+                    false
+            end;
+        _ -> false
+    end.
+
+sendfile_maybe_uncork(S, true) ->
+    _ = setopts(S, [{nopush,false}]),
+    ok;
+sendfile_maybe_uncork(_, false) ->
+    ok.
+
+sendfile_1(S, FileHandle, Offset, 0) ->
+    sendfile_1(S, FileHandle, Offset, (1 bsl 63) - 1);
+sendfile_1(_S, _FileHandle, Offset, Length) when
+         Offset < 0; Offset > ((1 bsl 63) - 1);
+         Length < 0; Length > ((1 bsl 63) - 1) ->
+    {error, einval};
+sendfile_1(S, FileHandle, Offset, Length) ->
+    Args = [FileHandle,
+            ?int64(Offset),
+            ?int64(Length)],
+    case ctl_cmd(S, ?TCP_REQ_SENDFILE, Args) of
+        {ok, []} ->
+            receive
+                {sendfile, S, {ok, SentLow, SentHigh}} ->
+                    {ok, SentLow bor (SentHigh bsl 32)};
+                {sendfile, S, {error, Reason}} ->
+                    {error, Reason};
+                {'EXIT', S, _Reason} ->
+                    {error, closed}
+            end;
+        {error, Reason} ->
+            {error, Reason}
+    end.
+
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%
 %% RECV(insock(), Length, [Timeout]) -> {ok,Data} | {error, Reason}
@@ -484,29 +722,46 @@ async_recv(S, Length, Time) ->
 %% oriented: preserved here only for API compatibility.
 %%
 recvfrom(S, Length) ->
-    recvfrom0(S, Length, -1).
+    recvfrom(S, Length, infinity).
 
-recvfrom(S, Length, infinity) ->
+recvfrom(S, Length, infinity) when is_port(S) ->
     recvfrom0(S, Length, -1);
-recvfrom(S, Length, Time) when is_integer(Time), Time < 16#ffffffff ->
-    recvfrom0(S, Length, Time);
-recvfrom(_, _, _) -> {error,einval}.
+recvfrom(S, Length, Time) when is_port(S) ->
+    if
+	is_integer(Time), 0 =< Time, Time < 16#ffffffff ->
+	    recvfrom0(S, Length, Time);
+	true ->
+	    {error, einval}
+    end.
 
 recvfrom0(S, Length, Time)
-  when is_port(S), is_integer(Length), Length >= 0, Length =< 16#ffffffff ->
+  when is_integer(Length), 0 =< Length, Length =< 16#ffffffff ->
     case ctl_cmd(S, ?PACKET_REQ_RECV,[enc_time(Time),?int32(Length)]) of
 	{ok,[R1,R0]} ->
 	    Ref = ?u16(R1,R0),
 	    receive
 		% Success, UDP:
-		{inet_async, S, Ref, {ok, [F,P1,P0 | AddrData]}} ->
-		    {IP,Data} = get_ip(F, AddrData),
-		    {ok, {IP, ?u16(P1,P0), Data}};
+		{inet_async, S, Ref, {ok, {[F | AddrData], AncData}}} ->
+                    %% With ancillary data
+		    case get_addr(F, AddrData) of
+			{{Family, _} = Addr, Data} when is_atom(Family) ->
+			    {ok, {Addr, 0, AncData, Data}};
+			{{IP, Port}, Data} ->
+			    {ok, {IP, Port, AncData, Data}}
+		    end;
+		{inet_async, S, Ref, {ok, [F | AddrData]}} ->
+                    %% Without ancillary data
+		    case get_addr(F, AddrData) of
+			{{Family, _} = Addr, Data} when is_atom(Family) ->
+			    {ok, {Addr, 0, Data}};
+			{{IP, Port}, Data} ->
+			    {ok, {IP, Port, Data}}
+		    end;
 
 		% Success, SCTP:
 		{inet_async, S, Ref, {ok, {[F,P1,P0 | Addr], AncData, DE}}} ->
-		    {IP, _}   = get_ip(F, Addr),
-		    {ok, {IP, ?u16(P1,P0), AncData, DE}};
+		    {IP, _} = get_ip(F, Addr),
+		    {ok, {IP, ?u16(P1, P0), AncData, DE}};
 
 		% Back-end error:
 		{inet_async, S, Ref, Error={error, _}} ->
@@ -525,21 +780,26 @@ recvfrom0(_, _, _) -> {error,einval}.
 
 peername(S) when is_port(S) ->
     case ctl_cmd(S, ?INET_REQ_PEER, []) of
-	{ok, [F, P1,P0 | Addr]} ->
-	    {IP, _} = get_ip(F, Addr),
-	    {ok, { IP, ?u16(P1, P0) }};
-	{error,_}=Error -> Error
+	{ok, [F | Addr]} ->
+	    {A, _} = get_addr(F, Addr),
+	    {ok, A};
+	{error, _} = Error -> Error
     end.
 
-setpeername(S, {IP,Port}) when is_port(S) ->
-    case ctl_cmd(S, ?INET_REQ_SETPEER, [?int16(Port),ip_to_bytes(IP)]) of
-	{ok,[]} -> ok;
-	{error,_}=Error -> Error
-    end;
 setpeername(S, undefined) when is_port(S) ->
     case ctl_cmd(S, ?INET_REQ_SETPEER, []) of
-	{ok,[]} -> ok;
-	{error,_}=Error -> Error
+	{ok, []} -> ok;
+	{error, _} = Error -> Error
+    end;
+setpeername(S, Addr) when is_port(S) ->
+    case type_value(set, addr, Addr) of
+	true ->
+	    case ctl_cmd(S, ?INET_REQ_SETPEER, enc_value(set, addr, Addr)) of
+		{ok, []} -> ok;
+		{error, _} = Error -> Error
+	    end;
+	false ->
+	    {error, einval}
     end.
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -580,21 +840,28 @@ peernames(S, AssocId)
 
 sockname(S) when is_port(S) ->
     case ctl_cmd(S, ?INET_REQ_NAME, []) of
-	{ok, [F, P1, P0 | Addr]} ->
-	    {IP, _} = get_ip(F, Addr),
-	    {ok, { IP, ?u16(P1, P0) }};
-	{error,_}=Error -> Error
+	{ok, [F | Addr]} ->
+	    {A, _} = get_addr(F, Addr),
+	    {ok, A};
+	{error, _} = Error -> Error
     end.
 
-setsockname(S, {IP,Port}) when is_port(S) ->
-    case ctl_cmd(S, ?INET_REQ_SETNAME, [?int16(Port),ip_to_bytes(IP)]) of
-	{ok,[]} -> ok;
-	{error,_}=Error -> Error
-    end;
 setsockname(S, undefined) when is_port(S) ->
     case ctl_cmd(S, ?INET_REQ_SETNAME, []) of
-	{ok,[]} -> ok;
-	{error,_}=Error -> Error
+	{ok, []} -> ok;
+	{error, _} = Error -> Error
+    end;
+setsockname(S, Addr) when is_port(S) ->
+    case type_value(set, addr, Addr) of
+	true ->
+	    case
+		ctl_cmd(S, ?INET_REQ_SETNAME, enc_value(set, addr, Addr))
+	    of
+		{ok, []} -> ok;
+		{error, _} = Error -> Error
+	    end;
+	false ->
+	    {error, einval}
     end.
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -724,9 +991,9 @@ chgopts(S, Opts) when is_port(S), is_list(Opts) ->
 
 getifaddrs(S) when is_port(S) ->
     case ctl_cmd(S, ?INET_REQ_GETIFADDRS, []) of
-	{ok, Data} ->
-	    {ok, comp_ifaddrs(build_ifaddrs(Data), ktree_empty())};
-	{error,enotsup} ->
+        {ok, Data} ->
+            {ok, comp_ifaddrs(build_ifaddrs(Data))};
+        {error,enotsup} ->
 	    case getiflist(S) of
 		{ok, IFs} ->
 		    {ok, getifaddrs_ifget(S, IFs)};
@@ -735,30 +1002,75 @@ getifaddrs(S) when is_port(S) ->
 	Err2 -> Err2
     end.
 
-%% Restructure interface properties per interface and remove duplicates
+%% Restructure interface properties per interface
 
-comp_ifaddrs([{If,Opts}|IfOpts], T) ->
-    case ktree_is_defined(If, T) of
-	true ->
-	    OptSet = comp_ifaddrs_add(ktree_get(If, T), Opts),
-	    comp_ifaddrs(IfOpts, ktree_update(If, OptSet, T));
-	false ->
-	    OptSet = comp_ifaddrs_add(ktree_empty(), Opts),
-	    comp_ifaddrs(IfOpts, ktree_insert(If, OptSet, T))
-     end;
-comp_ifaddrs([], T) ->
-    [{If,ktree_keys(ktree_get(If, T))} || If <- ktree_keys(T)].
-
-comp_ifaddrs_add(OptSet, [Opt|Opts]) ->
-    case ktree_is_defined(Opt, OptSet) of
-	true
-	  when element(1, Opt) =:= flags;
-	       element(1, Opt) =:= hwaddr ->
-	    comp_ifaddrs_add(OptSet, Opts);
-	_ ->
-	    comp_ifaddrs_add(ktree_insert(Opt, undefined, OptSet), Opts)
+comp_ifaddrs(IfOpts) ->
+    comp_ifaddrs(IfOpts, ktree_empty()).
+%%
+comp_ifaddrs([{If,[{flags,Flags}|Opts]}|IfOpts], IfT) ->
+    case ktree_is_defined(If, IfT) of
+        true ->
+            comp_ifaddrs(
+              IfOpts,
+              ktree_update(
+                If,
+                comp_ifaddrs_flags(Flags, Opts, ktree_get(If, IfT)),
+                IfT));
+        false ->
+            comp_ifaddrs(
+              IfOpts,
+              ktree_insert(
+                If,
+                comp_ifaddrs_flags(Flags, Opts, ktree_empty()),
+                IfT))
     end;
-comp_ifaddrs_add(OptSet, []) -> OptSet.
+comp_ifaddrs([], IfT) ->
+    comp_ifaddrs_2(ktree_keys(IfT), IfT).
+
+comp_ifaddrs_flags(Flags, Opts, FlagsT) ->
+    case ktree_is_defined(Flags, FlagsT) of
+        true ->
+            ktree_update(
+              Flags,
+              rev(Opts, ktree_get(Flags, FlagsT)),
+              FlagsT);
+        false ->
+            ktree_insert(Flags, rev(Opts), FlagsT)
+    end.
+
+comp_ifaddrs_2([If|Ifs], IfT) ->
+    FlagsT = ktree_get(If, IfT),
+    [{If,comp_ifaddrs_3(ktree_keys(FlagsT), FlagsT)}
+     | comp_ifaddrs_2(Ifs, IfT)];
+comp_ifaddrs_2([], _IfT) ->
+    [].
+%%
+comp_ifaddrs_3([Flags|FlagsL], FlagsT) ->
+    [{flags,Flags}|hwaddr_last(rev(ktree_get(Flags, FlagsT)))]
+        ++ hwaddr_last(comp_ifaddrs_3(FlagsL, FlagsT));
+comp_ifaddrs_3([], _FlagsT) ->
+    [].
+
+%% Place hwaddr last to look more like legacy emulation
+hwaddr_last(Opts) ->
+    hwaddr_last(Opts, Opts, []).
+%%
+hwaddr_last([{hwaddr,_} = Opt|Opts], L, R) ->
+    hwaddr_last(Opts, L, [Opt|R]);
+hwaddr_last([_|Opts], L, R) ->
+    hwaddr_last(Opts, L, R);
+hwaddr_last([], L, []) ->
+    L;
+hwaddr_last([], L, R) ->
+    rev(hwaddr_last(L, []), rev(R)).
+%%
+hwaddr_last([{hwaddr,_}|Opts], R) ->
+    hwaddr_last(Opts, R);
+hwaddr_last([Opt|Opts], R) ->
+    hwaddr_last(Opts, [Opt|R]);
+hwaddr_last([], R) ->
+    R.
+
 
 %% Legacy emulation of getifaddrs
 
@@ -766,21 +1078,19 @@ getifaddrs_ifget(_, []) -> [];
 getifaddrs_ifget(S, [IF|IFs]) ->
     case ifget(S, IF, [flags]) of
 	{ok,[{flags,Flags}]=FlagsVals} ->
-	    BroadOpts =
-		case member(broadcast, Flags) of
-		    true ->
-			[broadaddr,hwaddr];
-		    false ->
-			[hwaddr]
-		end,
-	    P2POpts =
-		case member(pointtopoint, Flags) of
-		    true ->
-			[dstaddr|BroadOpts];
-		    false ->
-			BroadOpts
-		end,
-	    getifaddrs_ifget(S, IFs, IF, FlagsVals, [addr,netmask|P2POpts]);
+            GetOpts =
+                case member(pointtopoint, Flags) of
+                    true ->
+                        [dstaddr,hwaddr];
+                    false ->
+                        case member(broadcast, Flags) of
+                            true ->
+                                [broadaddr,hwaddr];
+                            false ->
+                                [hwaddr]
+                        end
+                end,
+	    getifaddrs_ifget(S, IFs, IF, FlagsVals, [addr,netmask|GetOpts]);
 	_ ->
 	    getifaddrs_ifget(S, IFs, IF, [], [addr,netmask,hwaddr])
     end.
@@ -1122,7 +1432,14 @@ enc_opt(sndbuf)          -> ?INET_OPT_SNDBUF;
 enc_opt(recbuf)          -> ?INET_OPT_RCVBUF;
 enc_opt(priority)        -> ?INET_OPT_PRIORITY;
 enc_opt(tos)             -> ?INET_OPT_TOS;
+enc_opt(tclass)          -> ?INET_OPT_TCLASS;
+enc_opt(recvtos)         -> ?INET_OPT_RECVTOS;
+enc_opt(recvtclass)      -> ?INET_OPT_RECVTCLASS;
+enc_opt(pktoptions)      -> ?INET_OPT_PKTOPTIONS;
+enc_opt(ttl)             -> ?INET_OPT_TTL;
+enc_opt(recvttl)         -> ?INET_OPT_RECVTTL;
 enc_opt(nodelay)         -> ?TCP_OPT_NODELAY;
+enc_opt(nopush)          -> ?TCP_OPT_NOPUSH;
 enc_opt(multicast_if)    -> ?UDP_OPT_MULTICAST_IF;
 enc_opt(multicast_ttl)   -> ?UDP_OPT_MULTICAST_TTL;
 enc_opt(multicast_loop)  -> ?UDP_OPT_MULTICAST_LOOP;
@@ -1147,7 +1464,9 @@ enc_opt(packet_size)     -> ?INET_LOPT_PACKET_SIZE;
 enc_opt(read_packets)    -> ?INET_LOPT_READ_PACKETS;
 enc_opt(netns)           -> ?INET_LOPT_NETNS;
 enc_opt(show_econnreset) -> ?INET_LOPT_TCP_SHOW_ECONNRESET;
+enc_opt(line_delimiter)  -> ?INET_LOPT_LINE_DELIM;
 enc_opt(raw)             -> ?INET_OPT_RAW;
+enc_opt(bind_to_device)  -> ?INET_OPT_BIND_TO_DEVICE;
 % Names of SCTP opts:
 enc_opt(sctp_rtoinfo)	 	   -> ?SCTP_OPT_RTOINFO;
 enc_opt(sctp_associnfo)	 	   -> ?SCTP_OPT_ASSOCINFO;
@@ -1180,7 +1499,14 @@ dec_opt(?INET_OPT_SNDBUF)         -> sndbuf;
 dec_opt(?INET_OPT_RCVBUF)         -> recbuf;
 dec_opt(?INET_OPT_PRIORITY)       -> priority;
 dec_opt(?INET_OPT_TOS)            -> tos;
+dec_opt(?INET_OPT_TCLASS)         -> tclass;
 dec_opt(?TCP_OPT_NODELAY)         -> nodelay;
+dec_opt(?TCP_OPT_NOPUSH)          -> nopush;
+dec_opt(?INET_OPT_RECVTOS)        -> recvtos;
+dec_opt(?INET_OPT_RECVTCLASS)     -> recvtclass;
+dec_opt(?INET_OPT_PKTOPTIONS)     -> pktoptions;
+dec_opt(?INET_OPT_TTL)            -> ttl;
+dec_opt(?INET_OPT_RECVTTL)        -> recvttl;
 dec_opt(?UDP_OPT_MULTICAST_IF)    -> multicast_if;
 dec_opt(?UDP_OPT_MULTICAST_TTL)   -> multicast_ttl;
 dec_opt(?UDP_OPT_MULTICAST_LOOP)  -> multicast_loop;
@@ -1205,7 +1531,9 @@ dec_opt(?INET_LOPT_PACKET_SIZE)      -> packet_size;
 dec_opt(?INET_LOPT_READ_PACKETS)     -> read_packets;
 dec_opt(?INET_LOPT_NETNS)           -> netns;
 dec_opt(?INET_LOPT_TCP_SHOW_ECONNRESET) -> show_econnreset;
+dec_opt(?INET_LOPT_LINE_DELIM)      -> line_delimiter;
 dec_opt(?INET_OPT_RAW)              -> raw;
+dec_opt(?INET_OPT_BIND_TO_DEVICE) -> bind_to_device;
 dec_opt(I) when is_integer(I)     -> undefined.
 
 
@@ -1253,7 +1581,14 @@ type_opt_1(sndbuf)          -> int;
 type_opt_1(recbuf)          -> int;
 type_opt_1(priority)        -> int;
 type_opt_1(tos)             -> int;
+type_opt_1(tclass)          -> int;
+type_opt_1(recvtos)         -> bool;
+type_opt_1(recvtclass)      -> bool;
+type_opt_1(pktoptions)      -> opts;
+type_opt_1(ttl)             -> int;
+type_opt_1(recvttl)         -> bool;
 type_opt_1(nodelay)         -> bool;
+type_opt_1(nopush)          -> bool;
 type_opt_1(ipv6_v6only)     -> bool;
 %% multicast
 type_opt_1(multicast_ttl)   -> int;
@@ -1287,6 +1622,7 @@ type_opt_1(packet) ->
 	   {httph_bin,?TCP_PB_HTTPH_BIN},
 	   {ssl, ?TCP_PB_SSL_TLS}, % obsolete
 	   {ssl_tls, ?TCP_PB_SSL_TLS}]};
+type_opt_1(line_delimiter)  -> int;
 type_opt_1(mode) ->
     {enum,[{list, ?INET_MODE_LIST},
 	   {binary, ?INET_MODE_BINARY}]};
@@ -1305,6 +1641,7 @@ type_opt_1(packet_size)     -> uint;
 type_opt_1(read_packets)    -> uint;
 type_opt_1(netns)           -> binary;
 type_opt_1(show_econnreset) -> bool;
+type_opt_1(bind_to_device)  -> binary;
 %% 
 %% SCTP options (to be set). If the type is a record type, the corresponding
 %% record signature is returned, otherwise, an "elementary" type tag 
@@ -1415,7 +1752,7 @@ type_opt_1(O) when is_atom(O) -> undefined.
 
 %% Get. No supplied value.
 type_value(get, undefined)        -> false; % Undefined type
-%% These two clauses can not happen since they are only used
+%% These two clauses cannot happen since they are only used
 %% in record fields - from record fields they must have a
 %% value though it might be 'undefined', so record fields
 %% calls type_value/3, not type_value/2.
@@ -1493,14 +1830,49 @@ type_value_2(uint8, X)  when X band 16#ff =:= X       -> true;
 type_value_2(time, infinity)                          -> true;
 type_value_2(time, X) when is_integer(X), X >= 0      -> true;
 type_value_2(ip,{A,B,C,D}) when ?ip(A,B,C,D)          -> true;
+%%
 type_value_2(addr, {any,Port}) ->
     type_value_2(uint16, Port);
 type_value_2(addr, {loopback,Port}) ->
     type_value_2(uint16, Port);
-type_value_2(addr, {{A,B,C,D},Port}) when ?ip(A,B,C,D) ->
+type_value_2(addr, {IP,_} = Addr) when tuple_size(IP) =:= 4 ->
+    type_value_2(addr, {inet,Addr});
+type_value_2(addr, {IP,_} = Addr) when tuple_size(IP) =:= 8 ->
+    type_value_2(addr, {inet6,Addr});
+type_value_2(addr, {Local,_}) when is_list(Local); is_binary(Local) ->
+    type_value_2(addr, {local,Local});
+%%
+type_value_2(addr, {Family,{Tag,Port}})
+  when (Family =:= inet orelse Family =:= inet6) andalso
+       (Tag =:= any orelse Tag =:= loopback) ->
     type_value_2(uint16, Port);
-type_value_2(addr, {{A,B,C,D,E,F,G,H},Port}) when ?ip6(A,B,C,D,E,F,G,H) ->
+type_value_2(addr, {inet,{{A,B,C,D},Port}})
+  when ?ip(A,B,C,D) ->
     type_value_2(uint16, Port);
+type_value_2(addr, {inet6,{{A,B,C,D,E,F,G,H},Port}})
+  when ?ip6(A,B,C,D,E,F,G,H) ->
+    type_value_2(uint16, Port);
+type_value_2(addr, {local,Addr}) ->
+    if
+	is_binary(Addr) ->
+	    byte_size(Addr) =< 255;
+	true ->
+	    try
+		%% We either get a badarg from byte_size
+		%% or from characters_to_binary
+		byte_size(
+		  unicode:characters_to_binary(
+		    Addr, file:native_name_encoding()))
+	    of
+		N when N =< 255 ->
+		    true;
+		_ ->
+		    false
+	    catch error:badarg ->
+		    false
+	    end
+    end;
+%%
 type_value_2(ether,[X1,X2,X3,X4,X5,X6])
   when ?ether(X1,X2,X3,X4,X5,X6)                    -> true;
 type_value_2({enum,List}, Enum) -> 
@@ -1546,7 +1918,7 @@ type_value_2(_, _)         -> false.
 
 %% Get. No supplied value.
 %%
-%% These two clauses can not happen since they are only used
+%% These two clauses cannot happen since they are only used
 %% in record fields - from record fields they must have a
 %% value though it might be 'undefined', so record fields
 %% calls enc_value/3, not enc_value/2.
@@ -1608,6 +1980,7 @@ enc_value_2(time, Val)      -> ?int32(Val);
 enc_value_2(ip,{A,B,C,D})   -> [A,B,C,D];
 enc_value_2(ip, any)        -> [0,0,0,0];
 enc_value_2(ip, loopback)   -> [127,0,0,1];
+%%
 enc_value_2(addr, {any,Port}) ->
     [?INET_AF_ANY|?int16(Port)];
 enc_value_2(addr, {loopback,Port}) ->
@@ -1616,6 +1989,35 @@ enc_value_2(addr, {IP,Port}) when tuple_size(IP) =:= 4 ->
     [?INET_AF_INET,?int16(Port)|ip4_to_bytes(IP)];
 enc_value_2(addr, {IP,Port}) when tuple_size(IP) =:= 8 ->
     [?INET_AF_INET6,?int16(Port)|ip6_to_bytes(IP)];
+enc_value_2(addr, {File,_}) when is_list(File); is_binary(File) ->
+    [?INET_AF_LOCAL,iolist_size(File)|File];
+%%
+enc_value_2(addr, {inet,{any,Port}}) ->
+    [?INET_AF_INET,?int16(Port),0,0,0,0];
+enc_value_2(addr, {inet,{loopback,Port}}) ->
+    [?INET_AF_INET,?int16(Port),127,0,0,1];
+enc_value_2(addr, {inet,{IP,Port}}) ->
+    [?INET_AF_INET,?int16(Port)|ip4_to_bytes(IP)];
+enc_value_2(addr, {inet6,{any,Port}}) ->
+    [?INET_AF_INET6,?int16(Port),0,0,0,0,0,0,0,0];
+enc_value_2(addr, {inet6,{loopback,Port}}) ->
+    [?INET_AF_INET6,?int16(Port),0,0,0,0,0,0,0,1];
+enc_value_2(addr, {inet6,{IP,Port}}) ->
+    [?INET_AF_INET6,?int16(Port)|ip6_to_bytes(IP)];
+enc_value_2(addr, {local,Addr}) ->
+    %% A binary is passed as is, but anything else will be
+    %% regarded as a filename and therefore encoded according to
+    %% the current system filename encoding mode.
+    Bin =
+	if
+	    is_binary(Addr) ->
+		Addr;
+	    true ->
+		unicode:characters_to_binary(
+		  Addr, file:native_name_encoding())
+	end,
+    [?INET_AF_LOCAL,byte_size(Bin),Bin];
+%%
 enc_value_2(ether, [_,_,_,_,_,_]=Xs) -> Xs;
 enc_value_2(sockaddr, any) ->
     [?INET_AF_ANY];
@@ -1692,6 +2094,11 @@ dec_value(binary,[L0,L1,L2,L3|List]) ->
     Len = ?i32(L0,L1,L2,L3),
     {X,T}=split(Len,List),
     {list_to_binary(X),T};
+dec_value(opts, [L0,L1,L2,L3|List]) ->
+    Len = ?u32(L0,L1,L2,L3),
+    {X,T} = split(Len, List),
+    Opts = dec_opt_val(X),
+    {Opts,T};
 dec_value(Types, List) when is_tuple(Types) ->
     {L,T} = dec_value_tuple(Types, List, 1, []),
     {list_to_tuple(L),T};
@@ -2246,9 +2653,6 @@ utf8_to_characters(Bs, U, 0) ->
 utf8_to_characters([B|Bs], U, N) when ((B band 16#3F) bor 16#80) =:= B ->
     utf8_to_characters(Bs, (U bsl 6) bor (B band 16#3F), N-1).
 
-ip_to_bytes(IP) when tuple_size(IP) =:= 4 -> ip4_to_bytes(IP);
-ip_to_bytes(IP) when tuple_size(IP) =:= 8 -> ip6_to_bytes(IP).
-
 ip4_to_bytes({A,B,C,D}) ->
     [A band 16#ff, B band 16#ff, C band 16#ff, D band 16#ff].
 
@@ -2258,25 +2662,40 @@ ip6_to_bytes({A,B,C,D,E,F,G,H}) ->
 
 get_addrs([]) ->
     [];
-get_addrs([F,P1,P0|Addr]) ->
-    {IP,Addrs} = get_ip(F, Addr),
-    [{IP,?u16(P1, P0)}|get_addrs(Addrs)].
+get_addrs([F|Addrs]) ->
+    {Addr,Rest} = get_addr(F, Addrs),
+    [Addr|get_addrs(Rest)].
 
-get_ip(?INET_AF_INET, Addr)  -> get_ip4(Addr);
-get_ip(?INET_AF_INET6, Addr) -> get_ip6(Addr).
+get_addr(?INET_AF_LOCAL, [N|Addr]) ->
+    {A,Rest} = split(N, Addr),
+    {{local,iolist_to_binary(A)},Rest};
+get_addr(?INET_AF_UNSPEC, Rest) ->
+    {{unspec,<<>>},Rest};
+get_addr(?INET_AF_UNDEFINED, Rest) ->
+    {{undefined,<<>>},Rest};
+get_addr(Family, [P1,P0|Addr]) ->
+    {IP,Rest} = get_ip(Family, Addr),
+    {{IP,?u16(P1, P0)},Rest}.
+
+get_ip(?INET_AF_INET, Addr) ->
+    get_ip4(Addr);
+get_ip(?INET_AF_INET6, Addr) ->
+    get_ip6(Addr).
 
 get_ip4([A,B,C,D | T]) -> {{A,B,C,D},T}.
 
 get_ip6([X1,X2,X3,X4,X5,X6,X7,X8,X9,X10,X11,X12,X13,X14,X15,X16 | T]) ->
     { { ?u16(X1,X2),?u16(X3,X4),?u16(X5,X6),?u16(X7,X8),
-	?u16(X9,X10),?u16(X11,X12),?u16(X13,X14),?u16(X15,X16)}, T}.
+	?u16(X9,X10),?u16(X11,X12),?u16(X13,X14),?u16(X15,X16)},
+      T }.
 
+-define(ERTS_INET_DRV_CONTROL_MAGIC_NUMBER, 16#03f1a300).
 
 %% Control command
 ctl_cmd(Port, Cmd, Args) ->
     ?DBG_FORMAT("prim_inet:ctl_cmd(~p, ~p, ~p)~n", [Port,Cmd,Args]),
     Result =
-	try erlang:port_control(Port, Cmd, Args) of
+	try erlang:port_control(Port, Cmd+?ERTS_INET_DRV_CONTROL_MAGIC_NUMBER, Args) of
 	    [?INET_REP_OK|Reply]  -> {ok,Reply};
 	    [?INET_REP]  -> inet_reply;
 	    [?INET_REP_ERROR|Err] -> {error,list_to_atom(Err)}
